package demux.microsoft;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.channels.FileChannel;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.TreeSet;

import common.DirectMemoryManager;
import common.Job;
import common.JobStreamI;
import common.Logger;

public class WtvFileStream implements JobStreamI {
	private Logger logger=Logger.getLogger(this.getClass());
	private final String TABLE_ENTRIES=".entries.";
	
	public static final int META_CHUNK_SIZE=0x40000;
	public static final int META_SIZE=0x1000;
	
	private static final long META_CHUNK_MASK=~(META_CHUNK_SIZE - 1);
	
	private Job job;
	
	private File file;
	
	private FileChannel fileChannel;
	
	private long filePosition;
	
	private ByteBuffer buffer;
	
	private long streamSize;
	
	private TreeSet<Long> skipMajorChunks=new TreeSet<Long>();
	
	private int chunkSize=META_CHUNK_SIZE;
	
	private int readSize=20*1024*1024;
	
	private int skipSize;
	
	public WtvFileStream(Job job, File file) {
		this.job=job;
		this.file=file;
		reset();
	}
	@Override
	public void close() {
		if (fileChannel != null) {
			try {
				fileChannel.close();
			} catch (IOException e) {
			}
		}

	}

	@Override
	public Job getJob() {
		return job;
	}
	private void setJob(Job job) {
		job.setDescription(file.getAbsolutePath());
		FileChannel fileChannel=null;
		try {
			RandomAccessFile fileIn=new RandomAccessFile(file,"r"); 
			fileChannel=fileIn.getChannel();
			skipMajorChunks=new TreeSet<Long>();
			
			ByteBuffer fileLenBuffer=DirectMemoryManager.mapMemory(fileChannel, FileChannel.MapMode.READ_ONLY, 0x5c, 4);
			fileLenBuffer.order(ByteOrder.LITTLE_ENDIAN);
			int fileLen=fileLenBuffer.getInt(); //In meta units
			
			//Find the tables...
			byte guid[]=new byte[16];
			byte zero[]=new byte[16];
			ByteBuffer buffer=DirectMemoryManager.mapMemory(fileChannel, FileChannel.MapMode.READ_ONLY, 0x11000, 0x1000);
			buffer.order(ByteOrder.LITTLE_ENDIAN);
			buffer.get(guid);
			while (!Arrays.equals(guid, zero)) {
				int pos=buffer.position();
				int len=buffer.getInt();
				buffer.getInt();//len appears long;
				buffer.getLong();//Unknown
				int textLen=buffer.getInt();
				buffer.getInt();//textLen appears to be long
				StringBuffer sb=new StringBuffer();
				for (int i=0; i<textLen; i++) {
					char c=buffer.getChar();
					if (c != 0) {
						sb.append(c);
					}
				}
				String attr=sb.toString();
				if (attr.indexOf(TABLE_ENTRIES) > 0) {
					int value=buffer.getInt();
					int flag=buffer.getInt();
					logger.debug(attr + ": 0x" + Integer.toString(value, 16) + ": " + flag);
					if (flag==1) {
						addSkipChunks(fileChannel, value * META_SIZE, fileLen);
					} else {
						skipMajorChunks.add(getMetaChunkStart(value));
					}
				}
				
				buffer.position(len + pos - 16);
				buffer.get(guid);
			}
			
			StringBuffer sb=new StringBuffer("Skip Chunks: ");
			for (Long l : skipMajorChunks) {
				sb.append(Long.toHexString(l.longValue()));
				sb.append(", ");
			}
			logger.debug(sb);
			
			skipSize=(skipMajorChunks.size() * META_CHUNK_SIZE);
			streamSize=fileChannel.size() - skipSize;
			//We'll skip the first meta chunk out of the box
			skipMajorChunks.remove(0l);
			//Keep us from getting NPEs
			skipMajorChunks.add(fileChannel.size());
			fileChannel.close();
		} catch (IOException e) {
			e.printStackTrace();
			if (fileChannel != null) {
				try {
					fileChannel.close();
				} catch (IOException e1) {
				}
			}
		}
		setJobSize();
		job.setPosition(0);
	}
	
	/**
	 * Based on Meta tables info, figure out what where the META CHUNKS are
	 * Not perfect, but it should work
	 * @param fileChannel
	 * @param offset
	 * @throws IOException
	 */
	private void addSkipChunks(FileChannel fileChannel, long offset, int fileLen) throws IOException {
		ByteBuffer chunk=DirectMemoryManager.mapMemory(fileChannel, FileChannel.MapMode.READ_ONLY,  offset, META_CHUNK_SIZE);
		chunk.order(ByteOrder.LITTLE_ENDIAN);
		
		HashSet<Long> mySet=new HashSet<Long>();
		
		int i;
		while ((i=chunk.getInt()) != 0) {
			if (i > fileLen) {
				logger.error("Table entry greater than file len at: 0x" + Long.toString(offset, 16));
				return;
			}
			//Calculate the actual address and round off to the nearest major chunk start
			mySet.add(getMetaChunkStart(i));
		}
		skipMajorChunks.addAll(mySet);
	}

	private long getMetaChunkStart(int offset) {
		return (offset*(long)META_SIZE) & META_CHUNK_MASK;
	}
	public ByteBuffer getSizedBuffer(int ensuredBytes) throws IOException {
		ensureBuffer(ensuredBytes);
		return buffer;
	}
	@Override
	public long getStreamSize() {
		return streamSize;
	}

	@Override
	public boolean isEOS() throws IOException {
		return filePosition + buffer.position() - skipSize > streamSize  || job.getStatus()==Job.STATUS_ABORTED;
	}

	@Override
	public void open() throws IOException {
		if (fileChannel!=null) {
			close();
		}
		RandomAccessFile fileIn=new RandomAccessFile(file,"r"); 
		fileChannel=fileIn.getChannel();
		//Skip past the first META CHUNK
		filePosition=META_CHUNK_SIZE;
		readBuffer(false);
	}

	@Override
	public byte peek(int offset) throws IOException {
		ensureBuffer(offset);
		return buffer.get(buffer.position() + offset);
	}

	/**
	 * returns bytes or number of bytes left in file
	 * @param bytes
	 * @return
	 * @throws IOException
	 */
	private int ensureBuffer(int bytes) throws IOException {
		if (buffer.remaining() < bytes ) {
			ByteBuffer temp=ByteBuffer.allocateDirect(bytes);
			int neededBytes =bytes-buffer.remaining();
			temp.put(buffer);
			//Less efficient (from a memory standpoint), but logically less complex.  Skip Chunks are dealt with in one place
			neededBytes=Math.min(readBuffer(true), neededBytes);
			//Confusing, but it makes the buffer start before at the 0 point of the temp buffer
			filePosition -= temp.position();
			buffer.limit(neededBytes);
			temp.put(buffer);
			buffer=temp;
			buffer.position(0);
			return Math.min(buffer.remaining(), bytes);
		} else {
			return bytes;
		}
	}
	
	private int readBuffer(boolean advance) throws IOException {
		if (advance) {
			filePosition+=buffer.position();
		}
		while (filePosition==skipMajorChunks.first()) {
			filePosition+=META_CHUNK_SIZE;
			skipMajorChunks.remove(skipMajorChunks.first());
		}
		//For cut files, prevents us from moving beyond the end of the file
		if (filePosition > fileChannel.size()) {
			filePosition=fileChannel.size();
		}
		job.setPosition((int)(filePosition / Job.KB));
		long bytesToRead=Math.min(skipMajorChunks.first()-filePosition,Math.min(fileChannel.size() - filePosition, readSize));
		buffer=DirectMemoryManager.mapMemory(fileChannel, FileChannel.MapMode.READ_ONLY, filePosition, bytesToRead);
		return (int)bytesToRead;
	}
	
	@Override
	public void reset() {
		if (job!=null) {
			setJob(job);
		}
	}

	/**
	 * Not useful
	 */
	@Override
	public void setBufferChunks(int chunkSize, int numberOfChunks) {
		setChunkSize(chunkSize);
		readSize=chunkSize*numberOfChunks;
	}

	/**
	 * Not useful for WTV files
	 */
	@Override
	public void setChunkSize(int bytes) {
		this.chunkSize=bytes;
	}
	@Override
	public void setStreamSize(long size) {
		streamSize=size;
		setJobSize();
	}

	private void setJobSize() {
		job.setSize((int)(streamSize/Job.KB));
	}
	
	/**
	 * @return number of bytes skipped
	 */
	@Override
	public int skip(int bytes) throws IOException {
		int bufferSkip=Math.min(bytes, buffer.remaining());
		buffer.position(buffer.position() + bufferSkip);
		bytes -=bufferSkip;
		if (bytes > 0) {
			bytes=Math.min(bytes, readBuffer(true));
			buffer.position(bytes);
		}
		return bytes;
	}

	@Override
	public ByteBuffer slice(int len) throws IOException {
		len=ensureBuffer(len);
		
		int oldLimit=buffer.limit();
		buffer.limit(buffer.position()+len);
		ByteBuffer slice=buffer.slice();
		buffer.limit(oldLimit);
		buffer.position(buffer.position()+len);
		return slice;
	}
	public long getFilePos() throws IOException {
		return filePosition + buffer.position();
	}
}
