package muxer;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.nio.channels.WritableByteChannel;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import muxer.demux.PesDemuxerMuxer;
import muxer.demux.RawDemuxerMuxer;
import muxer.demux.VideoPesAudioRawDemuxerMuxer;
import muxer.pes.PesMuxer;
import muxer.ts.M2tsPesMuxer;
import muxer.ts.M2tsRawMuxer;
import muxer.ts.TsPesMuxer;
import muxer.ts.TsRawMuxer;

import common.BufferedWritableByteChannel;
import common.Job;
import common.JobStreamI;
import common.OutputStreamWritableByteChannel;
import common.StatusI;
import common.StreamDescriptor;
import common.packet.Packet;
import common.packet.PacketPart;
import common.packet.PesPacketPart;

import demux.Demuxer;
import demux.microsoft.WtvToPesDemuxer;
import demux.pes.DvdDemuxer;
import demux.ts.M2tsPesDemuxer;
import demux.ts.M2tsRawDemuxer;
import demux.ts.TsPesDemuxer;
import demux.ts.TsRawDemuxer;
import demux.ty.TyPesDemuxer;

public class MuxManager implements StatusI {
	public static int CONTAINER_PES=0;
	public static int CONTAINER_TS=1;
	public static int CONTAINER_M2TS=2;
	public static int CONTAINER_VOB;
	public static int CONTAINER_EVO;
	
	private static HashMap<String, RemuxInfo> extensionDemuxerMap=new HashMap<String, RemuxInfo>();
	
	public static class DemuxerMuxerInfo {
		private DemuxerMuxerInfo(Class demuxerClass, Class remuxerClass, String name, String ext) {
			this.demuxerClass=demuxerClass;
			this.remuxerClass=remuxerClass;
			this.name=name;
			this.ext=ext;
		}
		private Class demuxerClass;
		private Class remuxerClass;
		private String name;
		private String ext;
		public Class getDemuxerClass() {
			return demuxerClass;
		}
		public Class getRemuxerClass() {
			return remuxerClass;
		}
		public String getName() {
			return name;
		}
		public String getExt() {
			return ext;
		}
	}
	
	public static class RemuxInfo {
		private Class streamIdentiferClass;
		private DemuxerMuxerInfo[] demuxerMuxerInfo;
		
		private RemuxInfo(Class streamIdentiferClass, DemuxerMuxerInfo[] demuxerMuxerInfo) {
			this.streamIdentiferClass=streamIdentiferClass;
			this.demuxerMuxerInfo=demuxerMuxerInfo;
		}
	}
	
	static {
		RemuxInfo wtvInfo=new RemuxInfo(WtvToPesDemuxer.class, new DemuxerMuxerInfo[] {
			new DemuxerMuxerInfo(WtvToPesDemuxer.class, TsPesMuxer.class, "Transport Stream",".ts"),
			new DemuxerMuxerInfo(WtvToPesDemuxer.class, M2tsPesMuxer.class, "M2TS Stream",".m2ts"),
			new DemuxerMuxerInfo(WtvToPesDemuxer.class, PesMuxer.class, "MPG Stream",".mpg"),
			new DemuxerMuxerInfo(WtvToPesDemuxer.class, RawDemuxerMuxer.class, "Raw ES Streams",""),
			new DemuxerMuxerInfo(WtvToPesDemuxer.class, PesDemuxerMuxer.class, "Raw PES Streams",""),
			new DemuxerMuxerInfo(WtvToPesDemuxer.class, VideoPesAudioRawDemuxerMuxer.class, "PES Video&Raw Audio Streams","")
		});
		
		RemuxInfo tsInfo=new RemuxInfo(TsRawDemuxer.class, new DemuxerMuxerInfo[] {
			new DemuxerMuxerInfo(TsRawDemuxer.class, TsRawMuxer.class, "Transport Stream (raw)",".ts"),
			new DemuxerMuxerInfo(TsRawDemuxer.class, M2tsRawMuxer.class, "M2TS Stream (raw)",".m2ts"),
			new DemuxerMuxerInfo(TsPesDemuxer.class, TsPesMuxer.class, "Transport Stream",".ts"),
			new DemuxerMuxerInfo(TsPesDemuxer.class, M2tsPesMuxer.class, "M2TS Stream",".m2ts"),
			new DemuxerMuxerInfo(TsPesDemuxer.class, PesMuxer.class, "MPG Stream",".mpg"),
			new DemuxerMuxerInfo(TsPesDemuxer.class, RawDemuxerMuxer.class, "Raw ES Streams",""),
			new DemuxerMuxerInfo(TsPesDemuxer.class, PesDemuxerMuxer.class, "Raw PES Streams",""),
			new DemuxerMuxerInfo(TsPesDemuxer.class, VideoPesAudioRawDemuxerMuxer.class, "PES Video&Raw Audio Streams","")
		});
		
		RemuxInfo m2tsInfo=new RemuxInfo(M2tsRawDemuxer.class, new DemuxerMuxerInfo[] {
			new DemuxerMuxerInfo(M2tsRawDemuxer.class, TsRawMuxer.class, "Transport Stream (raw)",".ts"),
			new DemuxerMuxerInfo(M2tsRawDemuxer.class, M2tsRawMuxer.class, "M2TS Stream (raw)",".m2ts"),
			new DemuxerMuxerInfo(M2tsPesDemuxer.class, TsPesMuxer.class, "Transport Stream",".ts"),
			new DemuxerMuxerInfo(M2tsPesDemuxer.class, M2tsPesMuxer.class, "M2TS Stream",".m2ts"),
			new DemuxerMuxerInfo(M2tsPesDemuxer.class, PesMuxer.class, "MPG Stream",".mpg"),
			new DemuxerMuxerInfo(M2tsPesDemuxer.class, RawDemuxerMuxer.class, "ES Streams",""),
			new DemuxerMuxerInfo(M2tsPesDemuxer.class, PesDemuxerMuxer.class, "PES Streams",""),
			new DemuxerMuxerInfo(M2tsPesDemuxer.class, VideoPesAudioRawDemuxerMuxer.class, "PES Video & ES Audio Streams","")
		});
		
		RemuxInfo pesInfo=new RemuxInfo(DvdDemuxer.class, new DemuxerMuxerInfo[] {
			new DemuxerMuxerInfo(DvdDemuxer.class, TsPesMuxer.class, "Transport Stream",".ts"),
			new DemuxerMuxerInfo(DvdDemuxer.class, M2tsPesMuxer.class, "M2TS Stream",".m2ts"),
			new DemuxerMuxerInfo(DvdDemuxer.class, PesMuxer.class, "MPG Stream",".mpg"),
			new DemuxerMuxerInfo(DvdDemuxer.class, RawDemuxerMuxer.class, "ES Streams",""),
			new DemuxerMuxerInfo(DvdDemuxer.class, PesDemuxerMuxer.class, "PES Streams",""),
			new DemuxerMuxerInfo(DvdDemuxer.class, VideoPesAudioRawDemuxerMuxer.class, "PES Video & ES Audio Streams","")
		});
		RemuxInfo tyInfo=new RemuxInfo(TyPesDemuxer.class, new DemuxerMuxerInfo[] {
			new DemuxerMuxerInfo(TyPesDemuxer.class, PesMuxer.class, "MPG Stream",".mpg"),
			new DemuxerMuxerInfo(TyPesDemuxer.class, TsPesMuxer.class, "Transport Stream",".ts"),
			new DemuxerMuxerInfo(TyPesDemuxer.class, M2tsPesMuxer.class, "M2TS Stream",".m2ts"),
			new DemuxerMuxerInfo(TyPesDemuxer.class, RawDemuxerMuxer.class, "ES Streams",""),
			new DemuxerMuxerInfo(TyPesDemuxer.class, PesDemuxerMuxer.class, "PES Streams",""),
			new DemuxerMuxerInfo(TyPesDemuxer.class, VideoPesAudioRawDemuxerMuxer.class, "PES Video & ES Audio Streams","")
		});
		
		extensionDemuxerMap.put(".wtv", wtvInfo);
		extensionDemuxerMap.put(".ts", tsInfo);
		extensionDemuxerMap.put(".m2ts", m2tsInfo);
		extensionDemuxerMap.put(".mts", m2tsInfo);
		extensionDemuxerMap.put(".mpg", pesInfo);
		extensionDemuxerMap.put(".vob", pesInfo);
		extensionDemuxerMap.put(".evo", pesInfo);
		extensionDemuxerMap.put(".ty", tyInfo);
		extensionDemuxerMap.put(".ty+", tyInfo);
	}
	
	private Demuxer demuxer;
	private ArrayList<MuxFilter> muxerList=new ArrayList<MuxFilter>(5);
	private Map<Integer, StreamDescriptor> streamDescriptorMap;
	
	private static RemuxInfo getRemuxInfo(File file) {
		int endPos=file.getName().lastIndexOf('.');
		if (endPos > 0) {
			String extension=file.getName().substring(endPos).toLowerCase();
			return (RemuxInfo)extensionDemuxerMap.get(extension);
		} else {
			return null;
		}
	}
	
	/**
	 * Pick a demuxer based on file extension
	 * @param file
	 */
	public static Demuxer getStreamIdentiferDemuxer(File file) {
			RemuxInfo remuxInfo = getRemuxInfo(file);
			if (remuxInfo != null) {
				return getDemuxer(remuxInfo.streamIdentiferClass);
			}
		return null;
	}
	public static Demuxer getDemuxer(Class demuxer) {
		try {
			return (Demuxer)demuxer.newInstance();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return null;
	}
	public static DemuxerMuxerInfo[] getDemuxerRemuxerInfo(File file) {
		RemuxInfo remuxInfo = getRemuxInfo(file);
		return remuxInfo.demuxerMuxerInfo;
	}
	
	public static MuxFilter createRemuxer(Class remuxerClass, File file) throws IOException {
		Constructor<MuxFilter> constructor;
		Object params[]=new Object[1];
		try {
			constructor=remuxerClass.getConstructor(new Class[] {WritableByteChannel.class});
			params[0]=file.getName().equals("-") ? new OutputStreamWritableByteChannel(System.out) : new BufferedWritableByteChannel(file, 8*1024*1024);
		} catch (NoSuchMethodException nsme) {
			try {
				constructor=remuxerClass.getConstructor(new Class[] {String.class});
				params[0]=file.getAbsolutePath();
			} catch (NoSuchMethodException nsme2) {
				throw new RuntimeException("Could not find an OutputStream or String contructor for " + remuxerClass.getName());
			}
		}
		try {
			return (MuxFilter)constructor.newInstance(params);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		} 
	}
	
	public MuxManager(Demuxer demuxer) {
		this.demuxer=demuxer;
	}
	
	private void relink() {
		for (int i=0;i<muxerList.size();i++) {
			MuxFilter muxFilter=(MuxFilter)muxerList.get(i);
			muxFilter.setNextFilter((MuxFilter)(i+1==muxerList.size()?null:muxerList.get(i+1)));
			muxFilter.setMuxManager(this);
		}
	}
	public void add(MuxFilter muxer) throws IOException {
		muxerList.add(muxer);
		Collections.sort(muxerList);
		relink();
		muxer.init();
	}

	public boolean remove(MuxFilter muxer) {
		boolean b=muxerList.remove(muxer);
		relink();
		return b;
	}

	public Demuxer getDemuxer() {
		return demuxer;
	}

	public void setDemuxer(Demuxer demuxer) {
		this.demuxer = demuxer;
		this.muxerList=new ArrayList(5);
	}
	
	public PacketPart getPacketPart(int len) {
		return new PacketPart(len);
	}
	public PesPacketPart getPesPacketPart(int len) {
		return new PesPacketPart(len);
	}
	
	public int addPacket(Packet packet) throws IOException {
		return ((MuxFilter)muxerList.get(0)).addPacket(packet);
	}
	public void streamEnded() throws IOException {
		for (int i=0;i<muxerList.size();i++) {
			MuxFilter muxer=(MuxFilter)muxerList.get(i);
			muxer.streamEnd();
		}
	}
	public void mux(JobStreamI stream) throws IOException{
		stream.getJob().setStatus(Job.STATUS_RUNNING);
		demuxer.parse(stream, this);
		if (stream.getJob().getStatus()==Job.STATUS_RUNNING) {
			stream.getJob().setStatus(Job.STATUS_COMPLETE);
		}
	}
	public Map<Integer, StreamDescriptor> getStreamDescriptorMap(JobStreamI stream) throws IOException {
		if (streamDescriptorMap==null) {
			streamDescriptorMap=demuxer.getStreamDescriptorMap(stream, this);
		} 
		return streamDescriptorMap;
	}
	public PacketBuffer getPacketBuffer() {
		for (int i=0;i<muxerList.size();i++) {
			if (muxerList.get(i) instanceof PacketBuffer) {
				return (PacketBuffer)muxerList.get(i);
			}
		}
		return null;
	}
	public PacketBuffer createPacketBuffer() {
		PacketBuffer packetBuffer=getPacketBuffer();
		if (packetBuffer==null) {
			try {
				add(packetBuffer=new PacketBuffer());
			} catch (IOException e) {
				// Can't happen
				e.printStackTrace();
			}
		}
		return packetBuffer;
	}
	public static boolean isCopied(int pid, int allPids[]) {
		for (int i=0;i<allPids.length;i++) {
			if (pid==allPids[i]) {
				return true;
			}
		}
		return false;
	}
	public Map<Integer, StreamDescriptor> getMuxedStreamDescriptorMap() throws IOException {
		int streamIds[]=demuxer.getStreamIds();
		if (streamIds.length==0) {
			return streamDescriptorMap;
		}
		HashMap<Integer, StreamDescriptor> muxedMap=new HashMap<Integer, StreamDescriptor>();
		
		for (int i=0;i<streamIds.length;i++) {
			Integer key=Integer.valueOf(streamIds[i]);
			muxedMap.put(key, streamDescriptorMap.get(key));
		}
		return muxedMap;
	}
	
	/**
	 * 
	 * @return an order list of streams
	 * @throws IOException
	 */
	public Collection<StreamDescriptor> getMuxedStreamDescriptorColl() throws IOException {
		int streamIds[]=demuxer.getStreamIds();
		if (streamIds.length==0) {
			return streamDescriptorMap.values();
		}
		ArrayList<StreamDescriptor> list=new ArrayList<StreamDescriptor>(streamIds.length);
		
		for (int i=0;i<streamIds.length;i++) {
			list.add( streamDescriptorMap.get(streamIds[i]));
		}
		return list;
	}
}
